@@ -50,6 +50,13 @@ enum aml_obj_type {
 	AML_OBJ_IDLE,
 };
 
+struct aml_weak_ref {
+	void* obj;
+	LIST_ENTRY(aml_weak_ref) link;
+};
+
+LIST_HEAD(aml_weak_ref_list, aml_weak_ref);
+
 struct aml_obj {
 	enum aml_obj_type type;
 	int ref;
@@ -58,11 +65,11 @@ struct aml_obj {
 	aml_callback_fn cb;
 	unsigned long long id;
 	uint32_t n_events;
+	struct aml_weak_ref_list weak_refs;
 
 	void* backend_data;
 
 	LIST_ENTRY(aml_obj) link;
-	LIST_ENTRY(aml_obj) global_link;
 	TAILQ_ENTRY(aml_obj) event_link;
 };
 
@@ -82,8 +89,9 @@ struct aml_handler {
 struct aml_timer {
 	struct aml_obj obj;
 
-	uint32_t timeout;
+	uint64_t timeout;
 	uint64_t deadline;
+	bool expired;
 
 	LIST_ENTRY(aml_timer) link;
 };
@@ -136,9 +144,6 @@ struct aml {
 
 static struct aml* aml__default = NULL;
 
-static unsigned long long aml__obj_id = 0;
-static struct aml_obj_list aml__obj_list = LIST_HEAD_INITIALIZER(aml__obj_list);
-
 // TODO: Properly initialise this?
 static pthread_mutex_t aml__ref_mutex;
 
@@ -154,6 +159,8 @@ EXPORT const char aml_version[] = PROJECT_VERSION;
 EXPORT const char aml_version[] = "UNKNOWN";
 #endif
 
+EXPORT const int aml_unstable_abi_version = AML_UNSTABLE_API;
+
 EXPORT
 void aml_set_default(struct aml* aml)
 {
@@ -207,11 +214,11 @@ static void aml__dont_block(int fd)
 	fcntl(fd, F_SETFL, fcntl(fd, F_GETFL, 0) | O_NONBLOCK);
 }
 
-static uint64_t aml__gettime_ms(struct aml* self)
+static uint64_t aml__gettime_us(struct aml* self)
 {
 	struct timespec ts = { 0 };
 	clock_gettime(self->backend.clock, &ts);
-	return ts.tv_sec * 1000ULL + ts.tv_nsec / 1000000ULL;
+	return ts.tv_sec * UINT64_C(1000000) + ts.tv_nsec / UINT64_C(1000);
 }
 
 static void aml__ref_lock(void)
@@ -224,12 +231,45 @@ static void aml__ref_unlock(void)
 	pthread_mutex_unlock(&aml__ref_mutex);
 }
 
-static void aml__obj_global_ref(struct aml_obj* obj)
+EXPORT
+struct aml_weak_ref* aml_weak_ref_new(void* obj_ptr)
+{
+	struct aml_obj* obj = obj_ptr;
+
+	struct aml_weak_ref* self = calloc(1, sizeof(*self));
+	if (!self)
+		return NULL;
+
+	self->obj = obj;
+	aml__ref_lock();
+	LIST_INSERT_HEAD(&obj->weak_refs, self, link);
+	aml__ref_unlock();
+
+	return self;
+}
+
+EXPORT
+void aml_weak_ref_del(struct aml_weak_ref* self)
+{
+	if (!self)
+		return;
+
+	aml__ref_lock();
+	if (self->obj)
+		LIST_REMOVE(self, link);
+	aml__ref_unlock();
+	free(self);
+}
+
+EXPORT
+void* aml_weak_ref_read(struct aml_weak_ref* self)
 {
 	aml__ref_lock();
-	obj->id = aml__obj_id++;
-	LIST_INSERT_HEAD(&aml__obj_list, obj, global_link);
+	struct aml_obj* obj = self->obj;
+	if (obj)
+		obj->ref++;
 	aml__ref_unlock();
+	return obj;
 }
 
 static void on_self_pipe_read(void* obj) {
@@ -302,6 +342,7 @@ struct aml* aml_new(void)
 
 	self->obj.type = AML_OBJ_AML;
 	self->obj.ref = 1;
+	LIST_INIT(&self->obj.weak_refs);
 
 	LIST_INIT(&self->obj_list);
 	LIST_INIT(&self->timer_list);
@@ -328,8 +369,6 @@ struct aml* aml_new(void)
 	if (aml__init_self_pipe(self) < 0)
 		goto pipe_failure;
 
-	aml__obj_global_ref(&self->obj);
-
 	return self;
 
 pipe_failure:
@@ -374,17 +413,16 @@ struct aml_handler* aml_handler_new(int fd, aml_callback_fn callback,
 	self->obj.userdata = userdata;
 	self->obj.free_fn = free_fn;
 	self->obj.cb = callback;
+	LIST_INIT(&self->obj.weak_refs);
 
 	self->fd = fd;
 	self->event_mask = EVENT_MASK_DEFAULT;
 
-	aml__obj_global_ref(&self->obj);
-
 	return self;
 }
 
 EXPORT
-struct aml_timer* aml_timer_new(uint32_t timeout, aml_callback_fn callback,
+struct aml_timer* aml_timer_new(uint64_t timeout, aml_callback_fn callback,
                                 void* userdata, aml_free_fn free_fn)
 {
 	struct aml_timer* self = calloc(1, sizeof(*self));
@@ -396,16 +434,15 @@ struct aml_timer* aml_timer_new(uint32_t timeout, aml_callback_fn callback,
 	self->obj.userdata = userdata;
 	self->obj.free_fn = free_fn;
 	self->obj.cb = callback;
+	LIST_INIT(&self->obj.weak_refs);
 
 	self->timeout = timeout;
 
-	aml__obj_global_ref(&self->obj);
-
 	return self;
 }
 
 EXPORT
-struct aml_ticker* aml_ticker_new(uint32_t period, aml_callback_fn callback,
+struct aml_ticker* aml_ticker_new(uint64_t period, aml_callback_fn callback,
                                   void* userdata, aml_free_fn free_fn)
 {
 	struct aml_timer* timer =
@@ -427,11 +464,10 @@ struct aml_signal* aml_signal_new(int signo, aml_callback_fn callback,
 	self->obj.userdata = userdata;
 	self->obj.free_fn = free_fn;
 	self->obj.cb = callback;
+	LIST_INIT(&self->obj.weak_refs);
 
 	self->signo = signo;
 
-	aml__obj_global_ref(&self->obj);
-
 	return self;
 }
 
@@ -448,11 +484,10 @@ struct aml_work* aml_work_new(aml_callback_fn work_fn, aml_callback_fn callback,
 	self->obj.userdata = userdata;
 	self->obj.free_fn = free_fn;
 	self->obj.cb = callback;
+	LIST_INIT(&self->obj.weak_refs);
 
 	self->work_fn = work_fn;
 
-	aml__obj_global_ref(&self->obj);
-
 	return self;
 }
 
@@ -469,12 +504,23 @@ struct aml_idle* aml_idle_new(aml_callback_fn callback, void* userdata,
 	self->obj.userdata = userdata;
 	self->obj.free_fn = free_fn;
 	self->obj.cb = callback;
-
-	aml__obj_global_ref(&self->obj);
+	LIST_INIT(&self->obj.weak_refs);
 
 	return self;
 }
 
+static bool aml__obj_is_single_shot(void* ptr)
+{
+	struct aml_obj* obj = ptr;
+	switch (obj->type) {
+	case AML_OBJ_TIMER: /* fallthrough */
+	case AML_OBJ_WORK:
+		return true;
+	default:;
+	}
+	return false;
+}
+
 static bool aml__obj_is_started_unlocked(struct aml* self, void* obj)
 {
 	struct aml_obj* elem;
@@ -552,7 +598,8 @@ static int aml__start_handler(struct aml* self, struct aml_handler* handler)
 
 static int aml__start_timer(struct aml* self, struct aml_timer* timer)
 {
-	timer->deadline = aml__gettime_ms(self) + timer->timeout;
+	timer->deadline = aml__gettime_us(self) + timer->timeout;
+	timer->expired = false;
 
 	pthread_mutex_lock(&self->timer_list_mutex);
 	LIST_INSERT_HEAD(&self->timer_list, timer, link);
@@ -560,7 +607,6 @@ static int aml__start_timer(struct aml* self, struct aml_timer* timer)
 
 	if (timer->timeout == 0) {
 		assert(timer->obj.type != AML_OBJ_TICKER);
-		aml_stop(self, timer);
 		aml_emit(self, timer, 0);
 		aml_interrupt(self);
 		return 0;
@@ -646,7 +692,6 @@ static int aml__stop_signal(struct aml* self, struct aml_signal* sig)
 
 static int aml__stop_work(struct aml* self, struct aml_work* work)
 {
-	/* Note: The cb may be executed anyhow */
 	return 0;
 }
 
@@ -697,7 +742,7 @@ static struct aml_timer* aml__get_timer_with_earliest_deadline(struct aml* self)
 
 	pthread_mutex_lock(&self->timer_list_mutex);
 	LIST_FOREACH(timer, &self->timer_list, link)
-		if (timer->deadline < deadline) {
+		if (!timer->expired && timer->deadline < deadline) {
 			deadline = timer->deadline;
 			result = timer;
 		}
@@ -716,7 +761,7 @@ static bool aml__handle_timeout(struct aml* self, uint64_t now)
 
 	switch (timer->obj.type) {
 	case AML_OBJ_TIMER:
-		aml_stop(self, timer);
+		timer->expired = true;
 		break;
 	case AML_OBJ_TICKER:
 		timer->deadline += timer->timeout;
@@ -732,9 +777,10 @@ static bool aml__handle_timeout(struct aml* self, uint64_t now)
 static void aml__handle_idle(struct aml* self)
 {
 	struct aml_idle* idle;
+	struct aml_idle* tmp;
 
-	LIST_FOREACH(idle, &self->idle_list, link)
-		if (idle->obj.cb)
+	LIST_FOREACH_SAFE(idle, &self->idle_list, link, tmp)
+		if (idle->obj.cb && aml_is_started(self, idle))
 			idle->obj.cb(idle);
 }
 
@@ -745,8 +791,15 @@ static void aml__handle_event(struct aml* self, struct aml_obj* obj)
 	 */
 	aml_ref(obj);
 
-	if (obj->cb)
+	if (obj->cb && aml_is_started(self, obj)) {
+		/* Single-shot objects must be stopped before the callback so
+		 * that they can be restarted from within the callback.
+		 */
+		if (aml__obj_is_single_shot(obj))
+			aml_stop(self, obj);
+
 		obj->cb(obj);
+	}
 
 	if (obj->type == AML_OBJ_HANDLER) {
 		struct aml_handler* handler = (struct aml_handler*)obj;
@@ -761,9 +814,11 @@ static void aml__handle_event(struct aml* self, struct aml_obj* obj)
 
 /* Might exit earlier than timeout. It's up to the user to check */
 EXPORT
-int aml_poll(struct aml* self, int timeout)
+int aml_poll(struct aml* self, int64_t timeout_us)
 {
-	return aml__poll(self, timeout);
+	int timeout_ms = (timeout_us == INT64_C(-1))
+		? -1 : timeout_us / INT64_C(1000);
+	return aml__poll(self, timeout_ms);
 }
 
 static struct aml_obj* aml__event_dequeue(struct aml* self)
@@ -779,7 +834,7 @@ static struct aml_obj* aml__event_dequeue(struct aml* self)
 EXPORT
 void aml_dispatch(struct aml* self)
 {
-	uint64_t now = aml__gettime_ms(self);
+	uint64_t now = aml__gettime_us(self);
 	while (aml__handle_timeout(self, now));
 
 	struct aml_timer* earliest = aml__get_timer_with_earliest_deadline(self);
@@ -833,6 +888,7 @@ int aml_ref(void* obj)
 	struct aml_obj* self = obj;
 
 	aml__ref_lock();
+	assert(self->ref >= 0);
 	int ref = self->ref++;
 	aml__ref_unlock();
 
@@ -913,13 +969,20 @@ int aml_unref(void* obj)
 
 	aml__ref_lock();
 	int ref = --self->ref;
-	if (ref == 0)
-		LIST_REMOVE(self, global_link);
 	aml__ref_unlock();
+
 	assert(ref >= 0);
 	if (ref > 0)
 		goto done;
 
+	aml__ref_lock();
+	while (!LIST_EMPTY(&self->weak_refs)) {
+		struct aml_weak_ref* ref = LIST_FIRST(&self->weak_refs);
+		ref->obj = NULL;
+		LIST_REMOVE(ref, link);
+	}
+	aml__ref_unlock();
+
 	switch (self->type) {
 	case AML_OBJ_AML:
 		aml__free(obj);
@@ -950,32 +1013,6 @@ done:
 	return ref;
 }
 
-EXPORT
-unsigned long long aml_get_id(const void* obj)
-{
-	const struct aml_obj* aml_obj = obj;
-	return aml_obj->id;
-}
-
-EXPORT
-void* aml_try_ref(unsigned long long id)
-{
-	struct aml_obj* obj = NULL;
-
-	aml__ref_lock();
-	LIST_FOREACH(obj, &aml__obj_list, global_link)
-		if (obj->id == id)
-			break;
-
-	if (obj && obj->id == id)
-		obj->ref++;
-	else
-		obj = NULL;
-
-	aml__ref_unlock();
-	return obj;
-}
-
 EXPORT
 void* aml_get_userdata(const void* obj)
 {
@@ -1083,7 +1120,7 @@ void* aml_get_backend_state(const struct aml* self)
 }
 
 EXPORT
-void aml_set_duration(void* ptr, uint32_t duration)
+void aml_set_duration(void* ptr, uint64_t duration)
 {
 	struct aml_obj* obj = ptr;
 
